#!/usr/bin/env python3
"""
Created on Tue Jul  7 15:25:10 2020

@author: aimachine
"""
import argparse

import numpy as np


class lstm_config(argparse.Namespace):
    def __init__(
        self,
        npz_directory=None,
        npz_name=None,
        npz_val_name=None,
        key_categories=None,
        key_cord=None,
        stage_number=3,
        last_conv_factor=4,
        imagex=128,
        imagey=128,
        size_tminus=3,
        size_tplus=0,
        nboxes=1,
        depth=29,
        start_kernel=3,
        mid_kernel=3,
        startfilter=48,
        lstm_hidden_unit=16,
        epochs=100,
        learning_rate=1.0e-4,
        batch_size=10,
        model_name="NEATModel",
        yolo_v0=False,
        yolo_v1=True,
        yolo_v2=False,
        multievent=False,
        pure_lstm=False,
        show=True,
        **kwargs
    ):

        self.npz_directory = npz_directory
        self.npz_name = npz_name
        self.npz_val_name = npz_val_name
        self.key_categories = key_categories
        self.key_cord = key_cord
        self.yolo_v0 = yolo_v0
        self.yolo_v1 = yolo_v1
        self.yolo_v2 = yolo_v2
        self.nboxes = nboxes
        self.multievent = multievent
        self.categories = len(self.key_categories)
        self.box_vector = len(self.key_cord)
        self.depth = depth
        self.start_kernel = start_kernel
        self.mid_kernel = mid_kernel
        self.startfilter = startfilter
        self.lstm_hidden_unit = lstm_hidden_unit
        self.epochs = epochs
        self.stride = last_conv_factor
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.model_name = model_name
        self.show = show
        self.imagex = imagex
        self.imagey = imagey
        self.stage_number = stage_number
        self.last_conv_factor = last_conv_factor
        self.size_tminus = size_tminus
        self.size_tplus = size_tplus
        self.pure_lstm = pure_lstm
        self.is_valid()

    def to_json(self):

        config = {
            "npz_directory": self.npz_directory,
            "npz_name": self.npz_name,
            "npz_val_name": self.npz_val_name,
            "model_name": self.model_name,
            "multievent": self.multievent,
            "yolo_v0": self.yolo_v0,
            "yolo_v1": self.yolo_v1,
            "yolo_v2": self.yolo_v2,
            "nboxes": self.nboxes,
            "imagex": self.imagex,
            "imagey": self.imagey,
            "size_tminus": self.size_tminus,
            "size_tplus": self.size_tplus,
            "stride": self.stride,
            "depth": self.depth,
           
            "categories": self.categories,
            "box_vector": self.box_vector,
            "start_kernel": self.start_kernel,
            "mid_kernel": self.mid_kernel,
            "startfilter": self.startfilter,
            "lstm_hidden_unit": self.lstm_hidden_unit,
            "epochs": self.epochs,
            "learning_rate": self.learning_rate,
            "batch_size": self.batch_size,
            "show": self.show,
            "stage_number": self.stage_number,
            "last_conv_factor": self.last_conv_factor,
            "pure_lstm": self.pure_lstm,
        }

        for (k, v) in self.key_categories.items():
            config[k] = v

        for (k, v) in self.key_cord.items():
            config[k] = v

        return config

    def is_valid(self, return_invalid=False):
        """Check if configuration is valid.
        Returns
        -------
        bool
        Flag that indicates whether the current configuration values are valid.
        """

        def _is_int(v, low=None, high=None):
            return (
                isinstance(v, int)
                and (True if low is None else low <= v)
                and (True if high is None else v <= high)
            )

        ok = {}
        ok["npz_directory"] = isinstance(self.npz_directory, str)
        ok["npz_name"] = isinstance(self.npz_name, str)
        ok["npz_val_name"] = isinstance(self.npz_val_name, str)
        ok["pure_lstm"] = isinstance(self.pure_lstm, bool)
        ok["yolo_v0"] = isinstance(self.yolo_v0, bool)
        ok["yolo_v1"] = isinstance(self.yolo_v1, bool)
        ok["yolo_v2"] = isinstance(self.yolo_v2, bool)
        ok["depth"] = _is_int(self.depth, 1)
        
        ok["stride"] = _is_int(self.stride, 1)
        ok["start_kernel"] = _is_int(self.start_kernel, 1)
        ok["mid_kernel"] = _is_int(self.mid_kernel, 1)
        ok["startfilter"] = _is_int(self.startfilter, 1)
        ok["stage_number"] = _is_int(self.stage_number, 1)
        ok["last_conv_factor"] = _is_int(self.last_conv_factor, 1)
        ok["epochs"] = _is_int(self.epochs, 1)
        ok["nboxes"] = _is_int(self.nboxes, 1)

        ok["imagex"] = _is_int(self.imagex, 1)
        ok["imagey"] = _is_int(self.imagey, 1)
        ok["size_tminus"] = _is_int(self.size_tminus, 1)
        ok["size_tplus"] = _is_int(self.size_tplus, 1)

        ok["learning_rate"] = (
            np.isscalar(self.learning_rate) and self.learning_rate > 0
        )
        ok["multievent"] = isinstance(self.multievent, bool)
        ok["show"] = isinstance(self.show, bool)
        ok["categories"] = _is_int(len(self.key_categories), 1)
        ok["box_vector"] = _is_int(self.box_vector, 1)

        if return_invalid:
            return all(ok.values()), tuple(k for (k, v) in ok.items() if not v)
        else:
            return all(ok.values())


class volume_config(argparse.Namespace):
    def __init__(
        self,
        npz_directory=None,
        npz_name=None,
        npz_val_name=None,
        pure_lstm=False,
        key_categories=None,
        key_cord=None,
        stage_number=3,
        last_conv_factor=4,
        imagex=64,
        imagey=64,
        imagez=4,
        size_tminus=1,
        size_tplus=1,
        nboxes=1,
        depth= {'depth_0': 12, 'depth_1': 24, 'depth_2': 16},
        reduction = 0.5,
        weight_decay=1e-4,
        start_kernel=7,
        mid_kernel=3,
        startfilter=48,
        epochs=100,
        learning_rate=1.0e-4,
        batch_size=10,
        model_name="NEATModel",
        yolo_v0=False,
        yolo_v1=True,
        yolo_v2=False,
        multievent=False,
        show=True,
        **kwargs
    ):

        self.npz_directory = npz_directory
        self.npz_name = npz_name
        self.npz_val_name = npz_val_name
        self.key_categories = key_categories
        self.key_cord = key_cord
        self.yolo_v0 = yolo_v0
        self.yolo_v1 = yolo_v1
        self.yolo_v2 = yolo_v2
        self.nboxes = nboxes
        self.multievent = multievent
        self.categories = len(self.key_categories)
        self.box_vector = len(self.key_cord)
        self.depth = depth
        self.reduction = reduction
        self.weight_decay=weight_decay
        
        self.start_kernel = start_kernel
        self.mid_kernel = mid_kernel
        self.startfilter = startfilter
        self.epochs = epochs
        self.stride = last_conv_factor
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.model_name = model_name
        self.show = show
        self.imagex = imagex
        self.imagey = imagey
        self.imagez = imagez
        self.stage_number = stage_number
        self.last_conv_factor = last_conv_factor
        self.size_tminus = size_tminus
        self.size_tplus = size_tplus
        self.pure_lstm = pure_lstm
        self.is_valid()

    def to_json(self):

        config = {
            "npz_directory": self.npz_directory,
            "npz_name": self.npz_name,
            "npz_val_name": self.npz_val_name,
            "model_name": self.model_name,
            "multievent": self.multievent,
            "yolo_v0": self.yolo_v0,
            "yolo_v1": self.yolo_v1,
            "yolo_v2": self.yolo_v2,
            "nboxes": self.nboxes,
            "imagex": self.imagex,
            "imagey": self.imagey,
            "imagez": self.imagez,
            "size_tminus": self.size_tminus,
            "size_tplus": self.size_tplus,
            "stride": self.stride,
            "depth": self.depth,
            "reduction": self.reduction,
            "weight_decay": self.weight_decay,
            "categories": self.categories,
            "box_vector": self.box_vector,
            "start_kernel": self.start_kernel,
            "mid_kernel": self.mid_kernel,
            "startfilter": self.startfilter,
            "epochs": self.epochs,
            "learning_rate": self.learning_rate,
            "batch_size": self.batch_size,
            "show": self.show,
            "stage_number": self.stage_number,
            "last_conv_factor": self.last_conv_factor,
            "pure_lstm": self.pure_lstm,
        }

        for (k, v) in self.key_categories.items():
            config[k] = v

        for (k, v) in self.key_cord.items():
            config[k] = v

        return config

    def is_valid(self, return_invalid=False):
        """Check if configuration is valid.
        Returns
        -------
        bool
        Flag that indicates whether the current configuration values are valid.
        """

        def _is_int(v, low=None, high=None):
            return (
                isinstance(v, int)
                and (True if low is None else low <= v)
                and (True if high is None else v <= high)
            )

        ok = {}
        ok["npz_directory"] = isinstance(self.npz_directory, str)
        ok["npz_name"] = isinstance(self.npz_name, str)
        ok["npz_val_name"] = isinstance(self.npz_val_name, str)
        ok["yolo_v0"] = isinstance(self.yolo_v0, bool)
        ok["yolo_v1"] = isinstance(self.yolo_v1, bool)
        ok["yolo_v2"] = isinstance(self.yolo_v2, bool)
        ok["pure_lstm"] = isinstance(self.pure_lstm, bool)
        ok["depth"] = isinstance(self.pure_lstm, dict)
        ok["reduction"] = isinstance(self.reduction, float)
        ok["weight_decay"]=isinstance(self.weight_decay, float)
        ok["stride"] = _is_int(self.stride, 1)
        ok["start_kernel"] = _is_int(self.start_kernel, 1)
        ok["mid_kernel"] = _is_int(self.mid_kernel, 1)
        ok["startfilter"] = _is_int(self.startfilter, 1)
        ok["stage_number"] = _is_int(self.stage_number, 1)
        ok["last_conv_factor"] = _is_int(self.last_conv_factor, 1)
        ok["epochs"] = _is_int(self.epochs, 1)
        ok["nboxes"] = _is_int(self.nboxes, 1)

        ok["imagex"] = _is_int(self.imagex, 1)
        ok["imagey"] = _is_int(self.imagey, 1)
        ok["imagez"] = _is_int(self.imagez, 1)
        ok["size_tminus"] = _is_int(self.size_tminus, 1)
        ok["size_tplus"] = _is_int(self.size_tplus, 1)

        ok["learning_rate"] = (
            np.isscalar(self.learning_rate) and self.learning_rate > 0
        )
        ok["multievent"] = isinstance(self.multievent, bool)
        ok["show"] = isinstance(self.show, bool)
        ok["categories"] = _is_int(len(self.key_categories), 1)
        ok["box_vector"] = _is_int(self.box_vector, 1)

        if return_invalid:
            return all(ok.values()), tuple(k for (k, v) in ok.items() if not v)
        else:
            return all(ok.values())



import random

import keras
import matplotlib.pyplot as plt
import numpy as np
from keras import models
from matplotlib import cm


class PlotHistory(keras.callbacks.Callback):
    def __init__(
        self,
        Trainingmodel: models,
        X: np.ndarray,
        Y: np.ndarray,
        key_categories: dict,
        key_cord: dict,
        gridx: int,
        gridy: int,
        plot: bool = False,
        nboxes: int = 1,
    ):
        self.Trainingmodel = Trainingmodel
        self.X = X
        self.Y = Y
        self.plot = plot
        self.gridx = gridx
        self.gridy = gridy
        self.nboxes = nboxes
        self.key_cord = key_cord
        self.key_categories = key_categories

    def on_train_begin(self, logs={}):
        self.i = 0
        self.x = []
        self.losses = []
        self.val_losses = []
        self.acc = []
        self.val_acc = []
        if self.plot:
            self.fig = plt.figure()

        self.logs = []

    def on_epoch_end(self, epoch, logs={}):

        self.logs.append(logs)
        self.x.append(self.i)
        self.losses.append(logs.get("loss"))
        self.val_losses.append(logs.get("val_loss"))
        self.acc.append(logs.get("accuracy"))
        self.val_acc.append(logs.get("val_accuracy"))

        self.i += 1
        if self.plot:
            f, (ax1, ax2) = plt.subplots(1, 2, sharex=True)

            ax1.set_yscale("log")
            ax1.plot(self.x, self.losses, label="loss")
            ax1.plot(self.x, self.val_losses, label="val_loss")
            ax1.legend()

            ax2.plot(self.x, self.acc, label="accuracy")
            ax2.plot(self.x, self.val_acc, label="val_accuracy")
            ax2.legend()
            plt.show()
            # clear_output(True)
        idx = random.randint(1, self.X.shape[0] - 1)
        Printpredict(
            idx,
            self.Trainingmodel,
            self.X,
            self.Y,
            self.key_categories,
            self.key_cord,
            self.gridx,
            self.gridy,
            plot=self.plot,
            nboxes=self.nboxes,
        )


def Printpredict(
    idx,
    model,
    data,
    Truelabel,
    key_categories,
    key_cord,
    plot=False,
    nboxes=1,
):

    Image = data[idx]
    Truelabel = Truelabel[idx]
    data = np.expand_dims(Image, 0)
    prediction = model.predict(data)
    cols = 5

    if plot:
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, data.shape[1], figsize=(5 * cols, 5))
        fig.figsize = (20, 10)
    # The prediction vector is (1, categories + box_vector) dimensional, input data is (N T H W C) C is 1 in our case
    for j in range(0, data.shape[1]):

        img = Image[j, :, :, 0]
        if plot:
            ax[j].imshow(img, cm.Spectral)
    for i in range(0, prediction.shape[0]):

        try:
            maxevent = np.argmax(
                prediction[i, :, :, : len(key_categories)], axis=-1
            )
            trueevent = np.argmax(
                Truelabel[0, 0, : len(key_categories)], axis=-1
            )
        except ValueError:

            maxevent = np.argmax(
                prediction[i, :, :, :, : len(key_categories)], axis=-1
            )
            trueevent = np.argmax(
                Truelabel[0, 0, 0, : len(key_categories)], axis=-1
            )
        for (k, v) in key_categories.items():
            if v == maxevent:
                maxlabel = k
            if v == trueevent:
                truelabel = k
        try:
            print(
                "Predicted cell:",
                maxlabel,
                "Probability:",
                prediction[i, 0, 0, maxevent],
            )
            print("True Cell type:", truelabel)
        except ValueError:
            print(prediction[i, 0, 0, :])
            print("True Cell type:", truelabel)
            # print( "Predicted cell:", maxlabel , "Probability:", prediction[i,0,0,0,maxevent])
            # print('True Cell type:', truelabel)

        if nboxes > 1:
            for b in range(1, nboxes - 1):
                try:
                    prediction[
                        i,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] += prediction[
                        i,
                        :,
                        :,
                        len(key_categories)
                        + b * len(key_cord) : len(key_categories)
                        + (b + 1) * len(key_cord),
                    ]
                    prediction[
                        i,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] = prediction[
                        i,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] / (
                        nboxes - 1
                    )
                except ValueError:
                    prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] += prediction[
                        i,
                        :,
                        :,
                        len(key_categories)
                        + b * len(key_cord) : len(key_categories)
                        + (b + 1) * len(key_cord),
                    ]
                    prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] = prediction[
                        i,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] / (
                        nboxes - 1
                    )
        for (k, v) in key_cord.items():
            try:
                print(k, prediction[i, :, :, len(key_categories) + v])
                print(
                    "True positional value",
                    k,
                    Truelabel[0, 0, len(key_categories) + v],
                )
            except ValueError:

                print(k, prediction[i, :, :, :, len(key_categories) + v])
                print(
                    "True positional value",
                    k,
                    Truelabel[0, 0, 0, len(key_categories) + v],
                )

    if plot:
        plt.show()


class PlotSilverHistory(keras.callbacks.Callback):
    def __init__(
        self,
        Trainingmodel: models,
        X: np.ndarray,
        Y: np.ndarray,
        key_categories: dict,
        key_cord: dict,
        gridx: int,
        gridy: int,
        plot: bool = False,
        nboxes: int = 1,
    ):
        self.Trainingmodel = Trainingmodel
        self.X = X
        self.Y = Y
        self.plot = plot
        self.gridx = gridx
        self.gridy = gridy
        self.nboxes = nboxes
        self.key_cord = key_cord
        self.key_categories = key_categories

    def on_train_begin(self, logs={}):
        self.i = 0
        self.x = []
        self.losses = []
        self.val_losses = []
        self.acc = []
        self.val_acc = []
        if self.plot:
            self.fig = plt.figure()

        self.logs = []

    def on_epoch_end(self, epoch, logs={}):

        self.logs.append(logs)
        self.x.append(self.i)
        self.losses.append(logs.get("loss"))
        self.val_losses.append(logs.get("val_loss"))
        self.acc.append(logs.get("accuracy"))
        self.val_acc.append(logs.get("val_accuracy"))

        self.i += 1
        if self.plot:
            f, (ax1, ax2) = plt.subplots(1, 2, sharex=True)

            ax1.set_yscale("log")
            ax1.plot(self.x, self.losses, label="loss")
            ax1.plot(self.x, self.val_losses, label="val_loss")
            ax1.legend()

            ax2.plot(self.x, self.acc, label="accuracy")
            ax2.plot(self.x, self.val_acc, label="val_accuracy")
            ax2.legend()
            plt.show()
            # clear_output(True)
        idx = random.randint(1, self.X.shape[0] - 1)
        Printsilverpredict(
            idx,
            self.Trainingmodel,
            self.X,
            self.Y,
            self.key_categories,
            self.key_cord,
            self.gridx,
            self.gridy,
            plot=self.plot,
            nboxes=self.nboxes,
        )


def Printsilverpredict(
    idx,
    model,
    data,
    Truelabel,
    key_categories,
    key_cord,
    plot=False,
    nboxes=1,
):

    Image = data[idx]
    Truelabel = Truelabel[idx]
    data = np.expand_dims(Image, 0)
    prediction = model.predict(data)
    cols = 5

    if plot:
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, data.shape[1], figsize=(5 * cols, 5))
        fig.figsize = (20, 10)
    # The prediction vector is (1, categories + box_vector) dimensional, input data is (N H W T) C is T in our case
    for j in range(0, data.shape[-1]):

        img = Image[0, :, :, j]
        if plot:
            ax[j].imshow(img, cm.Spectral)
    for i in range(0, prediction.shape[0]):

        try:
            maxevent = np.argmax(
                prediction[i, :, :, : len(key_categories)], axis=-1
            )
            trueevent = np.argmax(
                Truelabel[0, 0, : len(key_categories)], axis=-1
            )
        except ValueError:

            maxevent = np.argmax(
                prediction[i, :, :, :, : len(key_categories)], axis=-1
            )
            trueevent = np.argmax(
                Truelabel[0, 0, 0, : len(key_categories)], axis=-1
            )
        for (k, v) in key_categories.items():
            if v == maxevent:
                maxlabel = k
            if v == trueevent:
                truelabel = k
        try:
            print(
                "Predicted cell:",
                maxlabel,
                "Probability:",
                prediction[i, 0, 0, maxevent],
            )
            print("True Cell type:", truelabel)
        except ValueError:
            print(prediction[i, 0, 0, :])
            print("True Cell type:", truelabel)
            # print( "Predicted cell:", maxlabel , "Probability:", prediction[i,0,0,0,maxevent])
            # print('True Cell type:', truelabel)

        if nboxes > 1:
            for b in range(1, nboxes - 1):
                try:
                    prediction[
                        i,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] += prediction[
                        i,
                        :,
                        :,
                        len(key_categories)
                        + b * len(key_cord) : len(key_categories)
                        + (b + 1) * len(key_cord),
                    ]
                    prediction[
                        i,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] = prediction[
                        i,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] / (
                        nboxes - 1
                    )
                except ValueError:
                    prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] += prediction[
                        i,
                        :,
                        :,
                        len(key_categories)
                        + b * len(key_cord) : len(key_categories)
                        + (b + 1) * len(key_cord),
                    ]
                    prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] = prediction[
                        i,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] / (
                        nboxes - 1
                    )
        for (k, v) in key_cord.items():
            try:
                print(k, prediction[i, :, :, len(key_categories) + v])
                print(
                    "True positional value",
                    k,
                    Truelabel[0, 0, len(key_categories) + v],
                )
            except ValueError:

                print(k, prediction[i, :, :, :, len(key_categories) + v])
                print(
                    "True positional value",
                    k,
                    Truelabel[0, 0, 0, len(key_categories) + v],
                )

    if plot:
        plt.show()


class PlotVolumeHistory(keras.callbacks.Callback):
    def __init__(
        self,
        Trainingmodel: models,
        X: np.ndarray,
        Y: np.ndarray,
        key_categories: dict,
        key_cord: dict,
        gridx: int,
        gridy: int,
        gridz: int,
        plot: bool = False,
        nboxes: int = 1,
    ):
        self.Trainingmodel = Trainingmodel
        self.X = X
        self.Y = Y
        self.plot = plot
        self.gridx = gridx
        self.gridy = gridy
        self.gridz = gridz
        self.nboxes = nboxes
        self.key_cord = key_cord
        self.key_categories = key_categories

    def on_train_begin(self, logs={}):
        self.i = 0
        self.x = []
        self.losses = []
        self.val_losses = []
        self.acc = []
        self.val_acc = []
        if self.plot:
            self.fig = plt.figure()

        self.logs = []

    def on_epoch_end(self,epoch,logs={}):

        self.logs.append(logs)
        self.x.append(self.i)
        self.losses.append(logs.get("loss"))
        self.val_losses.append(logs.get("val_loss"))
        self.acc.append(logs.get("accuracy"))
        self.val_acc.append(logs.get("val_accuracy"))

        self.i += 1
        if self.plot:
            f, (ax1, ax2) = plt.subplots(1, 2, sharex=True)

            ax1.set_yscale("log")
            ax1.plot(self.x, self.losses, label="loss")
            ax1.plot(self.x, self.val_losses, label="val_loss")
            ax1.legend()

            ax2.plot(self.x, self.acc, label="accuracy")
            ax2.plot(self.x, self.val_acc, label="val_accuracy")
            ax2.legend()
            plt.show()
            # clear_output(True)
        idx = random.randint(1, self.X.shape[0] - 1)
        PrintVolumepredict(
            idx,
            self.Trainingmodel,
            self.X,
            self.Y,
            self.key_categories,
            self.key_cord,
            plot=self.plot,
            nboxes=self.nboxes,
        )


def PrintVolumepredict(
    idx,
    model,
    data,
    Truelabel,
    key_categories,
    key_cord,
    plot=False,
    nboxes=1,
):

    Image = data[idx]
    Truelabel = Truelabel[idx]
    data = np.expand_dims(Image, 0)
    prediction = model.predict(data)
    cols = 5

    if plot:
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, data.shape[1], figsize=(5 * cols, 5))
        fig.figsize = (20, 10)
    # The prediction vector is (1, categories + box_vector) dimensional, input data is (N Z H W T) C is 1 in our case
    for j in range(0, data.shape[-1]):

        img = Image[Image.shape[0] // 2, :, :, j]
        if plot:
            ax[j].imshow(img, cm.Spectral)
    for i in range(0, prediction.shape[0]):

        try:
            maxevent = np.argmax(
                prediction[i, :, :, :, : len(key_categories)], axis=-1
            )
            trueevent = np.argmax(
                Truelabel[0, 0, 0, : len(key_categories)], axis=-1
            )
        except ValueError:

            maxevent = np.argmax(
                prediction[i, :, :, :, :, : len(key_categories)], axis=-1
            )
            trueevent = np.argmax(
                Truelabel[0, 0, 0, 0, : len(key_categories)], axis=-1
            )
        for (k, v) in key_categories.items():
            if v == maxevent:
                maxlabel = k
            if v == trueevent:
                truelabel = k
        try:
            print(
                "Predicted cell:",
                maxlabel,
                "Probability:",
                prediction[i, 0, 0, 0, maxevent],
            )
            print("True Cell type:", truelabel)
        except ValueError:
            print(prediction[i, 0, 0, 0, :])
            print("True Cell type:", truelabel)
            # print( "Predicted cell:", maxlabel , "Probability:", prediction[i,0,0,0,maxevent])
            # print('True Cell type:', truelabel)

        if nboxes > 1:
            for b in range(1, nboxes - 1):
                try:
                    prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] += prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories)
                        + b * len(key_cord) : len(key_categories)
                        + (b + 1) * len(key_cord),
                    ]
                    prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] = prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] / (
                        nboxes - 1
                    )
                except ValueError:
                    prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] += prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories)
                        + b * len(key_cord) : len(key_categories)
                        + (b + 1) * len(key_cord),
                    ]
                    prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] = prediction[
                        i,
                        :,
                        :,
                        :,
                        len(key_categories) : len(key_categories)
                        + len(key_cord),
                    ] / (
                        nboxes - 1
                    )
        for (k, v) in key_cord.items():
            try:
                print(k, prediction[i, :, :, :, len(key_categories) + v])
                print(
                    "True positional value",
                    k,
                    Truelabel[0, 0, 0, len(key_categories) + v],
                )
            except ValueError:

                print(k, prediction[i, :, :, :, :, len(key_categories) + v])
                print(
                    "True positional value",
                    k,
                    Truelabel[0, 0, 0, 0, len(key_categories) + v],
                )

    if plot:
        plt.show()


class PlotStaticHistory(keras.callbacks.Callback):
    def __init__(
        self,
        Trainingmodel: models,
        X: np.ndarray,
        Y: np.ndarray,
        key_categories: dict,
        key_cord: dict,
        gridx: int,
        gridy: int,
        plot: bool = False,
        nboxes: int = 1,
        class_only: bool = False,
    ):
        self.Trainingmodel = Trainingmodel
        self.X = X
        self.Y = Y
        self.gridx = gridx
        self.gridy = gridy
        self.plot = plot
        self.nboxes = nboxes
        self.key_cord = key_cord
        self.key_categories = key_categories
        self.class_only = class_only

    def on_train_begin(self, logs={}):
        self.i = 0
        self.x = []
        self.losses = []
        self.val_losses = []
        self.acc = []
        self.val_acc = []
        if self.plot:
            self.fig = plt.figure()

        self.logs = []

    def on_epoch_end(self, logs={}):
        print(logs)
        self.logs.append(logs)
        self.x.append(self.i)
        self.losses.append(logs.get("loss"))
        self.val_losses.append(logs.get("val_loss"))
        self.acc.append(logs.get("acc"))
        self.val_acc.append(logs.get("val_acc"))

        self.i += 1
        if self.plot:
            f, (ax1, ax2) = plt.subplots(1, 2, sharex=True)

            ax1.set_yscale("log")
            ax1.plot(self.x, self.losses, label="loss")
            ax1.plot(self.x, self.val_losses, label="val_loss")
            ax1.legend()
            ax2.plot(self.x, self.acc, label="acc")
            ax2.plot(self.x, self.val_acc, label="val_acc")
            ax2.legend()
            plt.show()
            # clear_output(True)
        idx = random.randint(1, self.X.shape[0] - 1)
        PrintStaticpredict(
            idx,
            self.Trainingmodel,
            self.X,
            self.Y,
            self.key_categories,
            self.key_cord,
            plot=self.plot,
            nboxes=self.nboxes,
            class_only=self.class_only,
        )


def PrintStaticpredict(
    idx,
    model: models,
    data: list,
    Truelabel: list,
    key_categories: dict,
    key_cord: dict,
    plot: bool = False,
    nboxes: int = 1,
    class_only: bool = False,
):

    Image = data[idx]
    Truelabel = Truelabel[idx]
    data = np.expand_dims(Image, 0)
    prediction = model.predict(data)

    # The prediction vector is (1, categories + box_vector) dimensional, input data is (N H W C) C is 1 in our case

    img = Image[:, :, 0]
    if plot:
        plt.imshow(img, cm.Spectral)
    for i in range(0, prediction.shape[0]):
        maxevent = np.argmax(
            prediction[i, :, :, : len(key_categories)], axis=-1
        )
        trueevent = np.argmax(Truelabel[0, 0, : len(key_categories)], axis=-1)
        for (k, v) in key_categories.items():
            if v == maxevent:
                maxlabel = k
            if v == trueevent:
                truelabel = k

        print(
            "Predicted cell:",
            maxlabel,
            "Probability:",
            prediction[i, 0, 0, maxevent],
        )
        print("True Cell type:", truelabel)
        if nboxes > 1:
            for b in range(1, nboxes - 1):
                prediction[
                    i,
                    :,
                    :,
                    len(key_categories) : len(key_categories) + len(key_cord),
                ] += prediction[
                    i,
                    :,
                    :,
                    len(key_categories)
                    + b * len(key_cord) : len(key_categories)
                    + (b + 1) * len(key_cord),
                ]
            prediction[
                i,
                :,
                :,
                len(key_categories) : len(key_categories) + len(key_cord),
            ] = prediction[
                i,
                :,
                :,
                len(key_categories) : len(key_categories) + len(key_cord),
            ] / (
                nboxes - 1
            )
        if class_only is False:
            for (k, v) in key_cord.items():

                print(k, prediction[i, :, :, len(key_categories) + v])
                print(
                    "True positional value",
                    k,
                    Truelabel[0, 0, len(key_categories) + v],
                )

    if plot:
        plt.show()
